#!/usr/bin/env python

# Copyright (C) 2002-2003
# Max-Planck-Institut fuer Radioastronomie Bonn
#
# Produced for the ALMA and APEX projects
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.  Correspondence concerning
# ALMA should be addressed as follows:
#
# Internet email: alma-sw-admin@nrao.edu
#
# Correspondence concerning APEX should be addressed as follows:
#
# Internet email: dmuders@mpifr-bonn.mpg.de
#
# DataparCoords.py
# module to take observed (Az,El) and fill in coordinates in MBFITS DATAPAR binary table
#    
#    Uses SLALIB by P.T.Wallace to do the hard work: observed-apparent-real coords
#    Uses WCSLIB by Mark Calabretta to do some more of the hard work:
#    astronomical basis frame -> user native spherical frame -> projection plane
#    Useful reading:
#    MBFITS definition document APEX-IFD-MPI-0002, in particular: 
#    - sect. 3 on coordinates (for description of WCS frames: astronomical basis frame, user native frame, intermediate pixel frame)
#    - the DATAPAR table
#    SLALIB documentation, http://www.starlink.rl.ac.uk/star/docs/sun67.htx/sun67.html:
#    - especially "Explanation and Examples: Celestial Coordinate Systems"
#    WCS documentation
#    - specifically Calabretta & Greisen 2002 A&A 395 1077 "Representation of celestial coordinates in FITS"
#
#    Methods are provided by class DCCreator.  See class documentation.
#
# Who			When		What
# J.Hatchell mpifr      24/02/04-        Initial version
#                       01/03/04         Skeleton from umbrello v.1.2
#                                        Az, El obs'd to RA, Dec J2000 real
#                                        wobbler offsets
#                                        B1950 and Galactic frames
#                       22/03/04         wcslib calls via dataparcoords.c
#                                        for user native spherical coords
#                                        and projection plane coords
#                       14/05/04         benchmarking
#                       20/05/04         speed up AzEltoRaDec2000 by
#                                        using SLA_QK methods on vectors
#                       25/05/04         TAI2UTC added to DCPosTime
#                                        interface object
#                       01/06/04         TAI2UTC and TAI2TT used instead of AstTime
#                       02/06/04         refraction correction is in antMount as of
#                                        Feb 2004 version.  Write code to use this.
#                       20/06/04         Full vector version
#                       26/07/04         Fixed LAST increment bug
#                       30/07/04         added AzElTrack method
# A.Belloche mpifr      05/08/04         put 2 printings inside an "if
#                                        config.DEBUG" statement in AzElTrack.
# A.Belloche mpifr      22/09/04         Replace dcinput.userframe.SourceAz by
#                                        dcinput.postime.SourceAz (was a bug,
#                                        idem for SourceEl). Add equinox =
#                                        dcinput.astframe.Equinox (l. ~830).
#                                        l. ~818 and ~828: use array max value
#                                        instead of first element.
#                                        Comment the UTC-TAI comment.
# D.Muders MPIfR        05/01/06         Updated TAI2UTC fallback value after
#					 introduction of new leap second on
#					 January 1st, 2006.
# to do:
# AzEltoDatapar using QCK methods -yes
# just refraction correction in azimuth output
# Test: look at (Ra,Dec,Az,El,UTC) in eg. JCMT data? OK
# PoletoSlopSbopSkop both ways (currently, just in reverse)
# Coordinate rotation and projection frame coords with WCSLIB - done
# refraction correction on elevation for (Az, El) output
# array rotation for wobbler chopping other than in azimuth - tricky
# parallactic angle
# (Az, El) output when array rotation is EQUA (ie. array offsets add in (RA, Dec))
# fix aoppat in py_slalib: takes one arg & should take 2
# - not needed, only updates LST.  Positional accuracy with d(LST) = d(UT)*1.0027379 is < 0.01''
# in an hour.  Weather changes affect refraction much more than this.
# rewrite c wrapper basistouserframe to take arrays
# polar motions
# proper motions

from math import *
from numarray import *
import copy
import time # for benchmarking

import slalib # P.Wallace (STARLINK) astronomical transformations library
#import wcslib # M.Calabretta (ATNF) world coordinate systems library
import dchelper # jh c helper functions to call wcslib

# import AstTime # now just uses UT1UTC and TAI2UTC and TAI2TT
from mbfits import MBFITSError
#from acspy import acslog
import CalibratorLog
import config
import profile

class DCPosTime:
    """ Fundamental positions and time """
    def __init__(self):
        self.SourceAz = array([-999.0]) # source obs'd az
        self.SourceEl = array([-999.0]) # source obs'd elevation
        self.SourceRA = array([-999.0]) # source obs'd azimuth
        self.SourceDec = array([-999.0]) # source Dec J2000 FK5
        self.Az=array([0.0])          # observed azimuths, N=0, E=90
        self.El=array([0.0])          # observed elevations
        self.MJD=array([0.0])         # timestamps: MJD in TIMESYS sys.(UTC or TAI)
        self.TimeSys="TAI"     # TIMESYS: TAI (for APEX) or UTC allowed
        self.UTC2UT1=0.0       # UT1-UTC in seconds.  Slalib DUT
        self.TAI2UTC=-33.0     # UTC-TAI in seconds; default -32 correct until
                               # next leap second. Only a fallback value in case
			       # nothing else is available.
        self.TAI2TT=+32.184    # TT-TAI conversion, default should be correct.
                               # for apparent to mean conversion
        self.CoordsWithWobbler=1 # Flag: coords include wobbler offset if true

    def __repr__(self):
        pass

class DCRefraction:
    """ Refraction correction data """
    def __init__(self):
        # for refraction correction, sensible defaults :
        self.Tamb=273.15          # Ambient temperature / K
        self.Patm=1013.25         # Atmospheric pressure / mB
        self.RelHumid=0.5         # Relative humidity / fraction
        self.LambdaEff=1000       # Effective wavelength / microns
                                  # no wavelength dependence in radio band
                                  # just set this greater than 100 microns
        self.TropLapseRate=0.0065 # Tropospheric Lapse Rate / K per m
        #or...
        self.Refraction=0.0       # available in antMount Feb 2004
                                  # use if non-zero
    def __repr__(self):
        pass

class DCSiteProperties:
    """ Observatory site properties """
    def __init__(self):
        # site properties
        self.SiteLong= -67.759222   # Observatory longitude / deg, east positive
        self.SiteLat= -23.00575    # Observatory latitude / deg
        self.SiteElev=5098.0    # Observatory elevation / m: APEX default
        self.PolarX=0.0      # polar correction < 0.3'': use 0 if unavailable
        self.PolarY=0.0      #  "  "
    def __repr__(self):
        pass

class DCAstFrame:
    """ Astronomical basis frame description """
    def __init__(self):
        # astronomical frame definition
        self.BasisLongSys="RA" # Basis system code, longitude, from CTYPE1
        self.BasisLatSys="DEC" # Basis system code, latitude, from CTYPE2
        self.RADecSys="FK5"  # RA/Dec system
        self.Equinox=2000.0 # Equinox

    def __repr__(self):
        pass
    
class DCUserFrame:
    """ User native frame description """
    def __init__(self):
        # for user frame offsets
        self.LongObj=0.0     # Object longitude in user native frame
        self.LatObj=0.0      # Object latitude in user native frame
        self.CRVAL1=array([0.0])      # User native frame zero in basis system long (usu. source pos.)
        self.CRVAL2=array([0.0])      # User native frame zero in basis system lat (usu. source pos.)
        self.LonPole=array([0.0])     # Native longitude of celestial pole
        self.LatPole=array([0.0])     # Basis latitude of native frame pole
#        self.FrameAngle=0.0  # rotation of user frame wrt basis frame at object position,
#                             # measured anticlockwise from line of latitude through object
                             # = ObsInp k or ticsAntMount I
        self.Proj="SFL"      # projection, last three characters of FITS CTYPE1/2
        
    def __repr__(self):
        pass

class DCDewarRot:
    """ Dewar rotation details"""
    def __init__(self):
        # for dewar rotation
        self.DewCabin = "Cassegrain"  # Location of bolometer
        self.DewRtMod="CABIN"     # Dewar rotation mode: CABIN/EQUA/HORIZ (FEBE header)
        self.DewarAngle=0.0   # Dewar angle/ deg (FEBE header)
    def __repr__(self):
        pass

class DCWobbler:
    """ Wobbler details"""
    def __init__(self):
        # for wobbler
        self.WobblerDispl=0.0 # Wobbler displacement / deg (MONITOR)
        self.WobblerDir="ALON" #Wobbler throw direction (SCAN header)
    def __repr__(self):
        pass

class SlopSbopSkop:
    """ ObsInp-style frame description  - see ObsInp manual section 10"""
    def __init__(self):
        self.SL0P=0.0 #longitude of origin of user native frame in ast. basis frame
        self.SB0P=0.0 #latiitude of origin of user native frame in ast. basis frame
        self.SK0P=0.0 # rotation which the positive beta axis makes with the basis meridian
    def __repr__(self):
        pass
    
class DCInput:
    """ All inputs needed to create a complete DATAPAR table.  Combines subclasses
    DCPosTime,DCAstFrame,DCRefraction,DCSiteProperties,DCUserFrame,DCDewarRot,DCWobbler
    """
    def __init__(self):
        self.postime=DCPosTime()
        self.astframe=DCAstFrame()
        self.refraction=DCRefraction()
        self.siteprops=DCSiteProperties()
        self.userframe=DCUserFrame()
        self.dewarrot=DCDewarRot()
        self.wobbler=DCWobbler()

    def __repr__(self):
        print self.postime
        print self.astframe
        print self.refraction
        print self.siteprops
        print self.userframe
        print self.dewarrot
        print self.wobbler

class DCReturn:

    """
    Return class for DataparCoordsCreator
    """

    def __init__(self):
        """
        """
        self.BasisLon=array([0.0])     # longitude in astronomical basis frame
        self.BasisLat=array([0.0])     # latitude " "
        self.Az=array([0.0])        # observed azimuth
        self.El=array([0.0])        # observed elevation
        self.CRPIX1=array([0.0])       # pixel location of user native frame zero
        self.CRPIX2=array([0.0])       # " "
        self.RotAngle=0.0     # rotation angle of array y axis wrt user frame vert. circ.
                              # measured +ve E of N
        self.PCmatrix=array([[0.0,0.0],[0.0,0.0]]) # array rotation matrix
        self.CoordsWithWobbler=1 # flag: coords include wobbler offset if true
        self.LongOff=array([0.0])      # long offset from source in user native frame
        self.LatOff=array([0.0])       # lat offset from source in user native frame

    def __repr__(self):
        """
        """
        pass


class DCCreator:

    """
    Coordinate conversion methods to complete the DATAPAR table from the timestamped observed Az and El and the information in the MONITOR, SCAN, & FEBE tables
    Uses SLALIB by P.T.Wallace to do (some of!) the hard work
    Uses WCSLIB by Mark Calabretta to do some more of the hard work

    Main method for filling in the complete DATAPAR section is AzEltoDatapar - see further documentation there

    uses CalibratorLog for logging

    """

    def __init__(self):
        """ Init: do nothing """
        self.arcsec = 1/3600.0 # 1 arcsec in degrees
        self.Logger=CalibratorLog.Logger()

        pass

    def AzElwithWobbler(self, postime, wobbler):
        if (wobbler.WobblerDir <> "ALON"):
            myerror = MBFITSError(msg="Wobbler switching only implemented for azimuth, not "+self.WobblerDir)
            raise myerror
        # these all should work with arrays
        azww = postime.Az + wobbler.WobblerDispl # add wobbler displacement to az
        azww = fmod(azww,360.0) # get in range
        elww = postime.El # no change in elevation
        # if wobbler is azimuth then it doesn't affect array rotation
        if config.DEBUG:
            print "DC: adding wobbler displacement of %4.2f deg in Az"%wobbler.WobblerDispl
#            print "(Az, El)obs along tel. axis:\t%10.7f\t%10.7f"%(postime.Az,postime.El)
            print "(Az, El)obs along tel. axis:", postime.Az, postime.El

#            print "(Az, El)obs with wobbler:\t%10.7f\t%10.7f"%(azww,elww)
            print "(Az, El)obs with wobbler:", azww, elww
            
        return(azww, elww)


    def AzEltoRaDec2000(self,postime,refraction,siteprops):
        """ Observed (Az, El) to mean (RA, Dec) FK5 J2000 via apparent (RA,Dec)
        inc. refraction, aberration, earth rotation, precession-nutation
        using SLA_OAP and SLA_AMP

        Arrays and refraction correction:
        Array orientation on the sky depends on the observed elevation
        (the actual telescope elevation), not the apparent elevation.
        The change in array shape/orientation with the refraction correction is very small
        array stretch here given as % stretch and arcsec error for an array 5' across
        el    stretch/%   arcsec err.
        30    0.12         0.2
        20    0.25         0.4
        10    0.91         1.4
        To take array stretch due to refraction into account, each array offset position would
        have to be calculated in (Az,El)obs then put through whole position transformation sequence
        individually.  Better: measure array offsets at elevation of observations.

        jh 19/5/04: Array version
        """

        # set up inputs to OAP, observed to apparent, checking inputs
        dut=postime.UTC2UT1
        # convert long, lat from degrees to radians
        elongm=siteprops.SiteLong*pi/180.0
        phim=siteprops.SiteLat*pi/180.0
        hm=siteprops.SiteElev
        xp=siteprops.PolarX
        yp=siteprops.PolarY
        tdk=refraction.Tamb
        pmb=refraction.Patm
        rh=refraction.RelHumid
        wl=refraction.LambdaEff
        tlr=refraction.TropLapseRate


        # start time to UTC, whatever the input
        if config.DEBUG:
            print "TimeSys = ", postime.TimeSys
        if (postime.TimeSys =="TAI"):            
            utcdate=postime.MJD+postime.TAI2UTC/(3600.0*24.0)
            tdbdate=postime.MJD[0]+postime.TAI2TT/(3600.0*24.0)
        elif (postime.TimeSys == "UTC"):
            # UTC: do nothing to utcdate
            utcdate=postime.MJD
            tdbdate = utcdate+(-postime.TAI2UTC+postime.TAI2TT)/(3600.0*24.0)
        else:
            print "mbfits.MBFITSError: ", err
            msg = "DC: cannot handle TimeSys "+postime.TimeSys
            myerror=MBFITSError(msg)
            raise(myerror)
        if config.DEBUG:
            print "MJDTAI, utcdate = ", postime.MJD[0], utcdate[0]

        # call sla_aoppa to set up mean to apparent parameters

        start = time.time()
        # get aoprms 
        aoprms = slalib.sla_aoppat(utcdate[0])
        aoprms=slalib.sla_aoppa(utcdate[0],dut,elongm,phim,hm,xp,yp,tdk,pmb,rh,wl,tlr,aoprms)
        if config.DEBUG:
             print "inputs to sla_aoppa:"
             print utcdate[0], dut,elongm,phim,hm,xp,yp,tdk,pmb,rh,wl,tlr

        # convert MJD array to UTC is in utcdate
        # check time range and split up as necessary
        # tbd

        # send array elements one by one to sla_oapqk
        type='A' # coords are az, zenith distance

        # output arrays
        #         rap=zeros(postime.Az.shape,Float64)
        #         dap=zeros(postime.El.shape,Float64)
        # multiply by 1.0 means that a new array is allocated
        rap=postime.Az*1.0
        dap=postime.El*1.0
        if config.DEBUG:
            print "postime.az, el = ", postime.Az, postime.El
        # oapqk loop through times/positions
        # this would be much quicker in c
        last0 = aoprms[13]
        for i in arange(len(postime.Az)):
            # update to time
            # this bit is fudged, because py_slalib has no 2nd arg
            # update aoprms: py_slalib call needs aoprms as 2nd arg!!!
            # aoprms= slalib.sla_aoppat(utcdate[i],aoprms)
            # approximate new last (radians):
            #last = aoprms[13]
            if config.DEBUG:
                print "old last = ", last0
            last = last0 + ((utcdate[i]-utcdate[0])*2*pi*1.0027379)
            # last seconds are shorter so d(last) > d(ut)
            # 1.0027379 approx. 365/364 taken from SLALIB gmst routine
            # Over an hour this approx. gives 0.01 arcsec additional error - negligible
            if config.DEBUG:
                print "new last = ", last
            # NB DUT only changes by 0.1-0.2 ms a day so that doesn't affect positional accuracy
            # over the length of an observation.
                
            aoprms[13] = last
            # quick position conversion
            ob1=postime.Az[i]*pi/180.0 #in radians
            ob2=(90.0-postime.El[i])*pi/180.0 # zenith distance in radians
            if config.DEBUG:
                print "ob1, ob2 = ", ob1, ob2
            (rap[i],dap[i])= slalib.sla_oapqk(type,ob1,ob2,aoprms)
            if config.DEBUG:
                print "rap,dap = ", rap[i], dap[i]

            # note: oapqk includes time-consuming refraction corr.
            # if this is known from antMount refraction then
            # this call could be replaced by a faster routine
            # see http://www.starlink.rl.ac.uk/star/docs/sun67.htx/node216.html
            # for the necessary steps

        duration = time.time() - start
        if config.DEBUG:            
            print "obs'd to apparent takes ", duration/len(postime.Az), " sec"


         
        if config.DEBUG:
            print "(RA, Dec)app:",rap*180.0/pi, dap*180.0/pi

#         if config.DEBUG:
#             # calculate refraction
#             # sla_refro returns zenith distance correction in radians
#             eps=pi/180/3600/10 # 0.1"
#             zz = slalib.sla_refro(ob2,hm,tdk,pmb,rh,wl,phim,tlr,eps)
#             p1 = ob2+zz
#             print "refraction corr. at ", postime.El, "(\") = ", zz*180/pi*3600
#             ob2plus = ob2+eps*100 # obs2 + 10''
#             zz = slalib.sla_refro(ob2plus,hm,tdk,pmb,rh,wl,phim,tlr,eps)
#             p2=ob2plus+zz
#             print "refraction corr. at ", postime.El, "(\") = ", zz*180/pi*3600
#             stretch = (p1-p2)*180*3600/pi/10.0*100
#             print "% stretch = ", stretch
#             # calculate hour angle with sla_dh2e
#             (ha,dec) = slalib.sla_dh2e(postime.Az*pi/180.0,postime.El*pi/180.0,phim)
#             print "HA, delta = ", ha*180.0/pi, dec*180.0/pi
#             # try polar motion
# #             xp=0.3/3600/180*pi
# #             yp=0.3/3600/180*pi
# #             (elong,phi,daz)=slalib.sla_polmo(elongm,phim,xp,yp)
# #             print "elong, phi, daz (\") = ", elong*180.0/pi, phi*180.0/pi, daz*180.0/pi*3600
#             # apparent-to-observed parameters
#             aoprms = array(zeros(14),type="d")
#             print aoprms
#             aoprms = slalib.sla_aoppa(utcdate,dut,elongm,phim,hm,xp,yp,tdk,pmb,rh,wl,tlr,aoprms)
#             print "aoprms = ", aoprms
#             print "LAST = ", aoprms[13]*12.0/pi
#             last=aoprms[13]*12.0/pi
#             lasth = floor(last)
#             lastm = floor((last-lasth)*60.0)
#             lasts = last-lasth-lastm/60.0
#            print "LAST = ", lasth, lastm, lasts
#            print "HA-LAST = ", ha*180.0/pi-aoprms[13]*180.0/pi

        # (RA, Dec) apparent to mean
        # set up parameters for first time in list
        # errors due to time difference within observation will be negligible        
        # as just proper motions and aberration
        # so get away with parameters in advance version
        # Adequate to use TT instead of TDB.  TDB-TT <= 0.0017s
        # AntMount just uses UTC instead of TDB, diff. 0.132s
        equinox=2000.0
        # ampqk loop
        # sla_amp apparent to mean
        amprms = slalib.sla_mappa(equinox,tdbdate)
        if config.DEBUG:
            print "tdbdate = ", tdbdate
        # output arrays
        rm = postime.Az*0.0
        dm = postime.El*0.0
        # could speed this loop up by doing in C
        for i in arange(len(rap)):
            (rm[i],dm[i]) = slalib.sla_ampqk(rap[i],dap[i],amprms)
        duration = time.time() - start
        if config.DEBUG:
            print "apparent to mean takes ", duration/len(postime.Az), " sec"

        return (rm*180.0/pi, dm*180.0/pi)


    def EqJ2000toBasisFrame(self, ra2000, dec2000, postime, astframe):
        """ J2000 (RA, Dec) to requested astronomical frame
        ra2000, dec2000: J2000 mean coordinates
        postime observation position and time object
        astframe requested astronomical frame object
        return: basislon, basislat"""

        # set up basislon,basislat arrays
        basislon = postime.Az
        basislat = postime.El

        if config.DEBUG:
            print "basislon, basislat = ", basislon, basislat

        # Processing depends on output astronomical basis frame
        if ((astframe.BasisLongSys=="ALON") & (astframe.BasisLatSys=="ALAT")):
            # Azimuth output, just refraction to correct
            # currently return input coords ie. no additional refraction corr.
            basislon=postime.Az
            basislat=postime.El
                
        elif ((astframe.BasisLongSys=="RA") & (astframe.BasisLatSys=="DEC")):
            if ((astframe.RADecSys=="FK5") & (astframe.Equinox == 2000.0)):
                basislon=ra2000
                basislat=dec2000
            elif ((astframe.RADecSys=="FK4") & (astframe.Equinox == 1950.0)):
                if config.DEBUG:
                    print "DC: converting to RA, dec 1950 (FK4)"
                try:
                    for i in arange(len(ra2000)):
                        (basislon[i],basislat[i],pma,pmd,p1950,v1950)=slalib.sla_fk524(ra2000[i]*pi/180.0,dec2000[i]*pi/180.0,0.0,0.0,0.0,0.0) # no proper motion or parallax
                except MBFITSError, err:
                    print "mbfits.MBFITSError: ", err
                    myerror=MBFITSError(msg="DC: sla_fk524 failed")
                    raise(myerror)
                    
                basislon=basislon*180.0/pi
                basislat=basislon*180.0/pi
            else:
                myerror = MBFITSError(msg="Equatorial system not J2000 FK5 or B1950 FK4")
                raise(myerror)
        elif ((astframe.BasisLongSys=="GLON") & (astframe.BasisLatSys=="GLAT")):
            if config.DEBUG:
                print "DC: converting to Galactic coordinates"
            try:
                for i in arange(len(ra2000)):
                    (basislon[i],basislat[i])=slalib.sla_eqgal(ra2000[i]*pi/180.0,dec2000[i]*pi/180.0)
            except MBFITSError, err:
                print "mbfits.MBFITSError: ", err
                raise(myerror)
            basislon=basislon*180.0/pi
            basislat=basislat*180.0/pi
        else:
            myerror= MBFITSError(msg="Coordinate system "+astframe.BasisLongSys+", "+astframe.BasisLatSys+" not yet implemented.")
            raise(myerror)

        return(basislon,basislat)

    def BasistoUserFrame(self,basislon,basislat, astframe, userframe):
        """
        calculate position in user native frame
        (theta, phi) native spherical coordinates and
        (x,y) intermediate pixel coords (projection plane coords)
        astframe = astronomical basis frame description
        userframe = DCUserFrame user frame description and coordinates

        use WCSLIB sphs2x/x2s to do coordinate rotation
        and WCSLIB prjs2x to do projection
        given frame centre and LONPOLE, LATPOLE

        inputs:
        astframe  AstFrame instance describing astronomical basis frame
        userframe UserFrame instance describing user native spherical frame
        return:
        (userlon, userlat) arrays of spherical coordinates in user native frame from frame origin
        (x,y) arrays of cartesian coordinates in user native frame from frame origin

        This method uses c helper functions in dataparcoords.c
        which call the wcslib methods directly
        
        @since 01/03/2004
        @author JH
        """

        # output arrays
        userlon = zeros(basislon.shape,Float64)
        userlat = zeros(basislat.shape,Float64)
        x = zeros(basislon.shape,Float64)
        y = zeros(basislat.shape,Float64)

        # call the c function from dchelper (dataparcoords.c)        
        for i in arange(len(basislon)):
            if config.DEBUG:
                print "Inputs to BasistoUserFrame: ", basislon[i],basislat[i],userframe.CRVAL1[i],userframe.CRVAL2[i],userframe.LonPole[i],userframe.LatPole[i],userframe.Proj
            (err,userlon[i],userlat[i],x[i],y[i]) = dchelper.BasistoUserFrame(basislon[i],basislat[i],userframe.CRVAL1[i],userframe.CRVAL2[i],userframe.LonPole[i],userframe.LatPole[i],userframe.Proj,0)
        if (err):
            myerror=MBFITSError(msg="DC: BasistoUserNative c helper function failed")
            raise(myerror)

        return(userlon, userlat, x, y)
        

    def UserOffsets(self, x,y, userframe):
        """
        calculate longoff and latoff offsets from source in projection of user native frame
        userframe = DCUserFrame user frame description and coordinates
        return:
        latoff, longoff = obs'd offsets from source in projection plane coords
        
        inputs:
        (x,y) position in user projection frame
        userframe UserFrame instance describing user native spherical frame (for source postion)
        
        @since 01/03/2004
        @author JH
        """

            
        longoff = x-userframe.LongObj
        latoff = y-userframe.LatObj

        return(longoff,latoff)
    

    def CalcArrayRotation(self, x,y,arrayx,arrayy):
        """
        Calculate PC matrix from new position of array (0,0) and (0,1)
        and pixel offsets of intermediate frame centre CRPIX1,CRPIX2
        
        Technique:
        take (0,1) vector in frame in which
        array fixed (cabin, ho or eq) add to array centre position,
        process through coordinate transforms, subtract array centre
        position, then
        PC = (b  a)   where (0,1) -> (a,b)
             (-a b)

        comments: For fields a degree across, the array rotation changes by about 2 deg
        The frame in which the array offsets are fixed is a spherical one so
        there is stretch/skew involved as well as rotation in the transfer to the user
        intermediate frame (x,y)
        strictly, positions should be calculated for individual array elements
        
        @since 19/04/2004
        @author JH
        """



        a = (arrayx - x)/self.arcsec
        b = (arrayy - y)/self.arcsec
        denom = sqrt(a*a+b*b)
        if config.DEBUG:
            print "(arrayx,arrayy):\t\t%10.7f\t%10.7f"%(arrayx,arrayy)
            print "(x,y): \t\t\t\t%10.7f\t%10.7f"%(x,y)
            print "(a,b): \t\t\t\t%10.7f\t%10.7f"%(a,b)
        pc = array([[b, a],[-a, b]])
        # offsets of intermediate frame from array centre in pixel coords
        CRPIX1 = b*-x+a*-y
        CRPIX2 = -a*-x+b*-y
        return (copy.deepcopy(pc),CRPIX1,CRPIX2)

    def CalcArrayRotAngle(self, userlon,userlat,arrayuserlon,arrayuserlat):
        """
        calculate angle of array in user frame
        replacement for CalcArrayRotation
        
        @since 18/05/2004
        @author JH
        """

        if (arrayuserlat-userlat == 0):
            if (arrayuserlon-userlon > 0):
                angle = 90.0
            elif (arrayuserlon-userlon < 0):
                angle = -90.0
            else:
                myerror = MBFITSError(msg="CalcArrayRotation: Faulty array (0,1) coords")
                raise myerror
        elif ((arrayuserlat-userlat) > 0):
            angle = 180.0/pi*arctan((arrayuserlon-userlon)/(arrayuserlat-userlat))
        elif ((arrayuserlat-userlat) < 0):
            angle = -180.0/pi*arctan((arrayuserlon-userlon)/(arrayuserlat-userlat))
            print "angle = ", angle
            addangle = 90.0
            addangle = self.copysign(addangle, angle)
            print "addangle = ", addangle
            angle = angle+addangle
            
        
        return (angle)

    def SlopSbopSkoptoPoles(self,sop):
        """ Convert OBSINP rotated frame description to MBFITS/WCS version
        for SBAS D scheme (not SBAS D * or SBAS P)
        see MBFITS document s.5.5.5
        SLOP/SBOP: Origin of user native frame in astronomical basis frame
        (usually the source position)
        SK0P: angle which user frame zero meridian makes with basis system meridian
        Return: (lonpole, latpole) Native longitude of celestial pole
        and Basis latitude of native frame pole
        jh 1/4/2004"""

        # see MBFITS s.5.5.5 for equations
        # OK for all quadrants?
        SK0P=fmod(sop.SK0P,180.0)*pi/180.0 # k
        SL0P=sop.SL0P*pi/180.0 # alpha_0
        SB0P=sop.SB0P*pi/180.0 # delta_0

        if config.DEBUG:
            print "\nx = ", sin(SB0P)
            print "y = ", sin(SK0P)*cos(SB0P)
            print "z = ", cos(SK0P)*cos(SB0P)
        
        # LONPOLE = longitude of basis pole in user native frame
        # LATPOLE = latitude of native pole in astronomical basis frame
        #         = latitude of basis pole in user native frame (by symmetry)
        # lonpole updated 3/3/2004 in s.5.5.5
        # tested for simple examples - OK
        if (SB0P==0):
            lonpole=-pi/2.0
        else:
            lonpole = arctan(sin(SK0P)*cos(SB0P)/sin(SB0P))
        lonpole=self.copysign(lonpole,sin(SK0P)*cos(SB0P))
        if (SB0P<0.0):
            lonpole=pi-lonpole

        latpole = arcsin(cos(SB0P)*cos(SK0P))
        latpole=self.copysign(latpole,cos(SK0P)*cos(SB0P))

        lonpole = lonpole*180.0/pi
        latpole = latpole*180.0/pi
        return (lonpole,latpole)
    
    def copysign(self,x,y):
        if (fabs(y)>0.0):
            return fabs(x)*y/fabs(y)
        else:
            return 0.0
    
    def PolestoSlopSbopSkop(self,userframe):
        """ Convert MBFITS/WCS user native frame description in terms of poles
        to OBSINP rotated frame description
        for SBAS D scheme (not SBAS D * or SBAS P)
        see MBFITS document s.5.5.5
        SLOP/SBOP: Origin of user native frame in astronomical basis frame
        (usually the source position)
        SK0P: angle which user frame zero meridian makes with basis system meridian
        userframe: includes userframe.LonPole Native longitude of celestial pole
        and userframe.LatPole Basis latitude of native frame pole

        Dont't think this is right yet.
        
        jh 1/4/2004"""

        # see MBFITS s.5.5.5 for equation
        # Conversion
        lonpole = userframe.LonPole*pi/180.0
        latpole = userframe.LatPole*pi/180.0
        # OK for all quadrants?
        SK0P=arctan(sin(CRVAL2)*tan(lonpole)/sin(latpole))
        SK0P=SK0P*180.0/pi
        SL0P=CRVAL1
        SB0P=CRVAL2
        return (SL0P,SB0P,SK0P)

    def CabinArray(self, postime, dewarrot):
        """ Initial array rotation for cassegrain/nasmyth cabins
        if dewar fixed in cabin frame
        Cassegrain: just dewar angle
        Nasmyth:    dewar angle & telescope observed elevation
        ( should strictly be elevation before pointing correction)
        postime: position and time object
        dewarrot: array dewar rotation and cabin description
        return (a,b): array (0,1) in horizontal (Az,El) frame
        """
        if (dewarrot.DewCabin == "Nasmyth"):
            # The obs'd elevation is the right one here
            # before the refraction correction to apparent
            # as it is the actual telescope elevation that determines the rotation on the sky
            paho = dewarrot.DewarAngle - postime.El
            a = -self.arcsec*sin(paho*pi/180.0)
            b = self.arcsec*cos(paho*pi/180.0)
        elif (dewarrot.DewCabin == "Cassegrain"):
            # APEX bolometers are going in cass cabin
            paho = dewarrot.DewarAngle
            a = -self.arcsec*sin(paho*pi/180.0)
            b = self.arcsec*cos(paho*pi/180.0)
# why did I have no dewar angle term before?            a = 0
#            b = self.arcsec
        return(a,b)

    def AzElTrack(self,dcinput):
        """ Calculate the lonpole and latpole to track a frame centred on the source but oriented parallel to (Az,El)
        inputs: dcinput structure with crval1,crval2 = source position
        optionally source ra, dec, az, el in postime
        return: dcinput.userframe """

        # copied from tryazoff3.py
        
        # simulate the (az,el) of the pointing source at this time
        # in fact this is available from AntMount

        dut=dcinput.postime.UTC2UT1
        elongm=dcinput.siteprops.SiteLong*pi/180.0
        phim=dcinput.siteprops.SiteLat*pi/180.0
        hm=dcinput.siteprops.SiteElev
        xp=dcinput.siteprops.PolarX
        yp=dcinput.siteprops.PolarY
        tdk=dcinput.refraction.Tamb
        pmb=dcinput.refraction.Patm
        rh=dcinput.refraction.RelHumid
        wl=dcinput.refraction.LambdaEff
        tlr=dcinput.refraction.TropLapseRate

        arrayshape=dcinput.postime.Az.shape
        arraysize = len(dcinput.postime.Az)

        if (dcinput.postime.TimeSys =="TAI"):            
            utcdate=dcinput.postime.MJD+dcinput.postime.TAI2UTC/(3600.0*24.0)
            tdbdate=dcinput.postime.MJD[0]+dcinput.postime.TAI2TT/(3600.0*24.0)
        elif (dcinput.postime.TimeSys == "UTC"):
            # UTC: do nothing to utcdate
            utcdate=dcinput.postime.MJD
            tdbdate = utcdate+(-dcinput.postime.TAI2UTC+dcinput.postime.TAI2TT)/(3600.0*24.0)
        else:
            print "mbfits.MBFITSError: ", err
            msg = "DC: cannot handle TimeSys "+dcinput.postime.TimeSys
            myerror=MBFITSError(msg)
            raise(myerror)

        if config.DEBUG:
            print "MJDTAI, utcdate = ", dcinput.postime.MJD[0], utcdate[0]

        if (max(dcinput.postime.SourceRA) > -999.0):
            sourcera=dcinput.postime.SourceRA
            sourcedec=dcinput.postime.SourceDec
        else:
            # Take RA and Dec of source as input CRVALs
            sourcera = dcinput.userframe.CRVAL1
            sourcedec = dcinput.userframe.CRVAL2
        if config.DEBUG:
            print "sourcera = ", sourcera
            print "sourcedec = ", sourcedec

        if (max(dcinput.postime.SourceAz) > -999.0):
            # these are az obs'd
            sourceaz = dcinput.postime.SourceAz
            sourceel = dcinput.postime.SourceEl
            equinox = dcinput.astframe.Equinox
        else:
            # convert source RA, Dec to Az, El
            rm=sourcera*pi/180.0
            dm=sourcedec*pi/180.0
            if config.DEBUG:
                print "rm, dm = ", rm, dm
            equinox = dcinput.astframe.Equinox
            rap = zeros(arrayshape, Float64)
            dap = zeros(arrayshape, Float64)
            aob = zeros(arrayshape, Float64)
            zob = zeros(arrayshape, Float64)
            for i in arange(arraysize):
                (rap[i],dap[i]) = slalib.sla_map(rm[i],dm[i],0,0,0,0,equinox,tdbdate)
                (aob[i],zob[i],hob,dob,rob)=slalib.sla_aop(rap[i],dap[i],utcdate[i],dut,elongm,phim,hm,xp,yp,tdk,pmb,rh,wl,tlr)
            sourceaz = aob*180.0/pi
            sourceel = 90.0-zob*180.0/pi

        if config.DEBUG:
            print "(AzElTrack: Az, El)obs source:",sourceaz,sourceel
        # take out the refraction correctionto get source (Az,El)app
        # elevation correction for refraction
        eps=pi/180/3600/10 # 0.1"
        zob = (90.0-sourceel)*pi/180.0
        zap = zeros(arrayshape,Float64)
        for i in arange(arraysize):
            zz = slalib.sla_refro(zob[i],hm,tdk,pmb,rh,wl,phim,tlr,eps)
            zap[i] = zob[i]+zz
        sourceazapp=sourceaz
        sourceelapp=90.0-zap*180.0/pi
        if config.DEBUG:
            print "AzElTrack:(Az, El)app source:",sourceazapp,sourceelapp
    
        # set up CRVALs and Lon,LatPole
        # frame is centred on (RA,Dec) of source
        # pole position calculation:
        # Get the celestial pole position in (az, el)observed
        # apparent az and z.d. of celestial pole
        # should really correct for polar motions but here quick & simple:
        if (dcinput.siteprops.SiteLat >= 0.0):        
            (aap,zap) = (0.0, 90.0-abs(dcinput.siteprops.SiteLat))
        else:
            (aap,zap) = (0.0, 90.0-dcinput.siteprops.SiteLat)
        
        #(rap,dap) = slalib.sla_map(rm,dm,0,0,0,0,equinox,utcdate)
        # obs'd az and el of celestial pole:
        # elevation correction for refraction
        # eps=pi/180/3600/10 # 0.1"
        # zz = slalib.sla_refro(zap*pi/180.0,hm,tdk,pmb,rh,wl,phim,tlr,eps)
        # zob = zap-zz*180.0/pi
        #elpole = 90.0-zob # obs'd el pole
        elpole = 90.0-zap
        azpole = aap # obs'd az pole
        if config.DEBUG:
            print "(Az, El)app pole:\t\t%10.7f\t%10.7f"%(azpole,elpole)
    
        # For comparison, do the same with slalib:
        rm=0.0
        dm=pi/2.0
        (rap,dap) = slalib.sla_map(rm,dm,0,0,0,0,equinox,tdbdate)
        # NB no refraction correction for apparent az,el
        nopmb=0.0
        (aob,zob,hob,dob,rob)=slalib.sla_aop(rap,dap,utcdate[0],dut,elongm,phim,hm,xp,yp,tdk,nopmb,rh,wl,tlr)
        # celestial zero in az and el
        azpole2 = aob*180.0/pi
        elpole2 = 90.0-zob*180.0/pi
        if config.DEBUG:
            print "(Az, El)app pole (slalib):\t%10.7f\t%10.7f"%(azpole2,elpole2)
#             # For comparison, do the same for (RA, Dec) = (0,0)
#         rap=0.0
#         dap=0.0
#         #(rap,dap) = slalib.sla_map(rm,dm,0,0,0,0,equinox,utcdate)
#         (aob,zob,hob,dob,rob)=slalib.sla_aop(rap,dap,utcdate,dut,elongm,phim,hm,xp,yp,tdk,pmb,rh,wl,tlr)
#         # celestial zero in az and el
#         az0 = aob*180.0/pi
#         el0 = 90.0-zob*180.0/pi
#         print "(Az, El)obs zero:\t\t%10.7f\t%10.7f"%(az0,el0)
    
        # swing from (az,el) to frame centred on source
        lonpole = zeros(arrayshape,Float64)
        latpole = zeros(arrayshape,Float64)
        for i in arange(arraysize):

            crval1 = sourceazapp[i]
            crval2 = sourceelapp[i]
            if (sourceelapp[i] >= 0.0):
                azazlonpole=0.0
            else:
                azazlonpole =180.0
            elellatpole =90.0-abs(sourceelapp[i])

            if config.DEBUG:
                print "AzElTrack -"
                print "Calling dchelper with:"
                print "crval1 = ", crval1
                print "crval2 = ", crval2
                print "lonpole = ", azazlonpole
                print "latpole = ", elellatpole
        
            (err,lonpole[i],latpole[i],x,y) = dchelper.BasistoUserFrame(azpole2,elpole2,crval1,crval2,azazlonpole,elellatpole,"CAR",0)
    
        # now we have the celestial pole position in the offset (az,el) sys
        if config.DEBUG:
            print "final lonpole,latpole:",lonpole,latpole
        dcinput.userframe.CRVAL1=sourcera # frame centre = source
        dcinput.userframe.CRVAL2=sourcedec 
        # here is the pole
        dcinput.userframe.LonPole=lonpole # user frame long of cel. pole
        # cel. frame lat. of user pole = user frame lat. of cel.pole
        dcinput.userframe.LatPole=-1.0*latpole
        return(dcinput.userframe)

############################################################

    def AzEltoDatapar(self,dcinput):
        """ Main method to complete all datapar parameters
       
        @param DataparCoordsInput DCInput : see class def. for details
        @return DataparCoordsReturn : see class def. for details
        @since 24/2/04
        @author JH
        """
        # check input time interval is not too long for quick version
        # 5 mins = rough guide
        
        #if ((dcinput.postime.MJD[-1]-dcinput.postime.MJD[0]) > 1.0/24.0):
        #    self.Logger.logger.logWarning('DataparCoords: expect position errors > 0.01 arcsec for MJD range > 5 mins')

        # initialise return class
        dcreturn = DCReturn()


        # Fill it in bit by bit


        # Include wobbler offsets or not depending on flag
        if (dcinput.postime.CoordsWithWobbler):
            try:
                start = time.time()
                (azww,elww) = self.AzElwithWobbler(dcinput.postime,dcinput.wobbler)
                duration = time.time() - start
                if config.DEBUG:
                    print "AzElwithWobbler takes ", duration, " sec"

            except MBFITSError, err:
                print "mbfits.MBFITSError: ", err
                myerror=MBFITSError(msg="DC: AzElwithWobbler failed")
                raise(myerror)
            dcinput.postime.Az = azww
            dcinput.postime.El = elww

        # output Az, El
        dcreturn.Az=dcinput.postime.Az
        dcreturn.El=dcinput.postime.El

        # Initial array rotation for cassegrain/nasmyth cabins if dewar fixed in cabin frame
        # (a,b) are updated to position of array (0,1) in each new frame
        # make a new postime instance with the array offset position
# leave out array rotation for now
#         arraypostime = copy.deepcopy(dcinput.postime)
#         if (dcinput.dewarrot.DewRtMod == "CABIN"):
#             try:
#                 (a,b) = self.CabinArray(dcinput.postime, dcinput.dewarrot)
#             except MBFITSError, err:
#                 print "mbfits.MBFITSError: ", err
#                 myerror=MBFITSError(msg="DC: CabinArray failed")
#                 raise(myerror)    
#         elif (dcinput.dewarrot.DewRtMod == "HORIZ"):
#             (a,b) = (0,self.arcsec)
#         elif (dcinput.dewarrot.DewRtMod == "EQUA"):
#             # equatorial case is handled later
#             (a,b) = (0,0)
#         arraypostime.Az = dcinput.postime.Az+a
#         arraypostime.El = dcinput.postime.El+b
#         if config.DEBUG:
#             print "(Az, El)obs for array (0,1):\t%10.7f\t%10.7f"%(arraypostime.Az,arraypostime.El)
                        
        # do conversion to mean place RA, Dec J2000 from obs'd Az, El
        try:
            #start=time.time()
            (ra2000,dec2000) =  self.AzEltoRaDec2000(dcinput.postime, dcinput.refraction,dcinput.siteprops)
            #duration=time.time()-start
            #print "azeltoradec2000 took ", duration, " s"

        except MBFITSError, err:
            print "mbfits.MBFITSError: ", err
            myerror=MBFITSError(msg="DC: AzEltoRaDec failed")
            raise(myerror)
        if config.DEBUG:
            print "(RA,Dec)2000:\t", ra2000, dec2000


        arraydec2000=dec2000
        arrayra2000=ra2000
        if (dcinput.dewarrot.DewRtMod == "EQUA"):
            (a,b) = (0,self.arcsec)
            (arrayra2000,arraydec2000) = (ra2000+a,dec2000+b)
#        if (config.DEBUG):
#            print "(RA,Dec)2000 + array off:\t%10.7f\t%10.7f"%(arrayra2000,arraydec2000)
            
        # conversion to user requested astronomical basis frame (B1950, Galactic etc.)
        try:
            #start = time.time()
            (basislon,basislat) = self.EqJ2000toBasisFrame(ra2000, dec2000, dcinput.postime, dcinput.astframe)
            #duration = time.time() - start
            #if config.DEBUG:
            #    print "EqJ2000toBasisFrame takes ", duration, " sec"

            #(arraybasislon,arraybasislat) = self.EqJ2000toBasisFrame(arrayra2000, arraydec2000, arraypostime, dcinput.astframe)
        except MBFITSError, err:
            print "mbfits.MBFITSError: ", err
            myerror=MBFITSError(msg="DC: EqJ2000toBasisFrame failed")
            raise(myerror)

        # fill in DATAPAR
        dcreturn.BasisLon=basislon
        dcreturn.BasisLat=basislat
        if config.DEBUG:
            print "(Lon, Lat)basis ",basislon,basislat

        # Next: rotation to user native spherical frame, projection, and subtraction of source position
        # to give LONGOFF, LATOFF
        try:
            start = time.time()
            (userlon,userlat,x,y) = self.BasistoUserFrame(basislon, basislat, dcinput.astframe,dcinput.userframe)
            duration = time.time() - start
            if config.DEBUG:
                print "BasistoUserFrame takes ", duration, " sec"
#            (arrayuserlon,arrayuserlat,arrayx,arrayy) = self.BasistoUserFrame(arraybasislon, arraybasislat, dcinput.astframe,dcinput.userframe)
        except MBFITSError, err:
            print "mbfits.MBFITSError: ", err
            myerror=MBFITSError(msg="DC: BasistoUserFrame failed")
            raise(myerror)
        if config.DEBUG:
            print "(Lon, Lat)user:",userlon,"\n",userlat
#            print "(Lon, Lat)user + array off:\t%10.7f\t%10.7f"%(arrayuserlon,arrayuserlat)
            print "(x, y)user:",x,"\n",y
#            print "(x, y)user + array off:\t\t%10.7f\t%10.7f"%(arrayx,arrayy)
            
        # offsets from source in user frame
        try:
            start = time.time()
            (longoff, latoff) = self.UserOffsets(x,y, dcinput.userframe)
            duration = time.time() - start
            if config.DEBUG:
                print "useroffsets takes ", duration, " sec"

        except MBFITSError, err:
            print "mbfits.MBFITSError: ", err
            myerror=MBFITSError(msg="DC: UserOffsets failed")
            raise(myerror)
        if config.DEBUG:
            print "(longobj, latobj)user:\t\t%10.7f\t%10.7f"%(dcinput.userframe.LongObj,dcinput.userframe.LatObj)
            print "(longoff, latoff) from object: ", longoff*3600,latoff*3600

        # As all FITS angles, LongOff and LatOff are in degrees
        dcreturn.LongOff = longoff
        dcreturn.LatOff = latoff

        # Finally: array rotation
        # pixel coordinates scrapped as array pixels do not define a regular grid image
        # and array (p1,p2) coordinate frame is distorted by projection wrt (x,y)
        # no simple offset/rotation to get from (p1,p2) to (x,y) so don't try and store it

        try:
            start = time.time()
            #dcreturn.RotAngle = self.CalcArrayRotAngle(userlon, userlat, arrayuserlon, arrayuserlat)
            # CalcArrayRotation defunct - PCmatrix, CRPIX1, CRPIX2 removed
            #(dcreturn.PCmatrix,dcreturn.CRPIX1,dcreturn.CRPIX2) = self.CalcArrayRotation(x,y,arrayx,arrayy)
            duration = time.time() - start
            if config.DEBUG:
                print "CalcArrayRotAngle takes ", duration, " sec"
        except MBFITSError, err:
            print "mbfits.MBFITSError: ", err
            myerror=MBFITSError(msg="DC: CalcArrayRotAngle failed")
            raise(myerror)
#        if config.DEBUG:
#            print "Array rotation angle in user frame = ", dcreturn.RotAngle
#             print "PC matrix = "
#             print dcreturn.PCmatrix
#             stretch = sqrt(dcreturn.PCmatrix[0,0]**2+dcreturn.PCmatrix[0,1]**2)
#             print "array stretch = ", stretch
#             print "array rotation in user frame / deg = ", arccos(dcreturn.PCmatrix[0,0]/stretch)*180.0/pi
#             print "(CRPIX1, CRPIX2):\t%10.7f\t%10.7f"%(dcreturn.CRPIX1,dcreturn.CRPIX2)
# 
#        return (dcreturn)
        return(copy.deepcopy(dcreturn))

############################################################

# Benchmarking 14/5/04:
# 200ms for first call: TAI-UTC in AstTime 0.153s the first time
#  40ms subsequent call
#  10-20ms AzEltoRADec (x2 for array)
# of which:
#   20 - 330 ms AstTime TAI to UTC
#   17ms slalib.sla_oap
# new, array version:
#   20 ms for each of 10 positions
# of which:
#   10-15 ms to do time conversion
#   0.4-2.8 ms obs'd to apparent excluding time change per pos for 10
#   0.4 ms apparent to mean excluding time change per pos for 10
#   0.5ms sla_oapqk loop per pos.
#   0.3ms slalib.sla_amp per pos.(why the big variation?)
#
#   3-4ms    CalcArrayRotation
# 0.4ms    AzElwithWobbler
# 8x10^-2ms BasistoUserFrame
# 4x10^-2ms EqJ2000toBasisFrame
# 2x10^-2ms UserOffsets

# 11 ms     SlopSbopSkoptoPoles

# Array rotation options
# PC matrix should be calculated for user native spherical frame not projection frame
# because array distorts - axes no longer perpendicular - away from frame centre
# eg. if user native frame is (RA, Dec) and observing away from (0,0) then
# array is no longer square / hexagonal in projection frame
#
# angle w.r.t. user native frame vertical circle is a valid option for storing array rotation
# is this worth calculating and storing in MBFITS?
#
# heterodyne arrays
# CLASS writer calculates a position for each array element
# 1) if array angle wrt user frame is stored in DATAPAR
# spherical frame to spherical frame conversion to get array element positions
# frame definitions: array angle, array zero in u.n.f. -> array centre, pole in u.n.f.
# convert using
#(err,userlon,userlat,x,y) = dchelper.BasistoUserFrame(basislon,basislat,userframe.CRVAL1,userframe.CRVAL2,userframe.LonPole,userframe.LatPole,userframe.Proj,0)
# The timings for this will be similar to BasistoUserFrame + SL0PSBOPSKOP
# 11.3 ms (slopsbopskop for each array position) + 0.08 ms for each pixel (vector call quicker) to get (theta, phi) and (x,y) 
#
# 2) if array angle wrt user frame is not stored in DATAPAR
# ClassWriter needs to call (AzEl ->) RaDec -> UserNative  for each array position.  Better to do this once and store the frame rotation parameters in DATAPAR.
#
# Could store array pole instead of rotation angle in DATAPAR if this is easier to calculate.
#
# What does continuum reduction require?
# Continuum observers (Frank) only use user native = basis RA, Dec frame so only parallactic angle needed = angle with user frame
#
# How does parallactic angle depend on refraction correction?
# Have to scrap PC matrix as array to (x,y) isn't simple rotation


# how to speed up
# do we need refraction correction?
# better TAI-UTC code (PTWallace?)
# quick refraction correction
# vectorise (effects of changing time on RA/Dec?)
# array rotation: remove or calc. angle by hand

# vectorise (Az,El) -> (RA, Dec):
# sky rotates at 15''/s
# for positional accuracy of p'', keep delta t to < p/15 s
# p     | delta t
# 0.1'' | 7 ms
# 0.01''| 0.7 ms




